import uuid
from typing import List, Literal
from pydantic import BaseModel, Field

class BaseCommand(BaseModel):
    id: str = Field(
        default_factory=lambda: str(uuid.uuid4()),
        description=(
            "The id of the saved step form"
        )
    )

class PauseCommand(BaseCommand):
    stepType: str = Field(default="pause", description="The type of the step")
    stepName: str = Field(default="暂停", description="Name of the step") 
    moduleId: str = Field(..., description="The ID of the module")
    pauseAction: Literal["untilTime", "untilTemperature", "untilResume"] = Field(..., description="Type of pause action")
    pauseSecond: int = Field(default=0, description="Number of seconds to pause")
    pauseMessage: str = Field(default="", description="Message to display during pause")
    pauseTemperature: int = Field(default=0, description="Temperature to pause until")

class HeaterShakerCommand(BaseCommand):
    stepType: str = Field(default="heaterShaker", description="The type of the step")
    stepName: str = Field(default="加热振荡", description="Name of the step")
    moduleId: str = Field(..., description="The ID of the heater-shaker module")
    setShake: bool = Field(default=False, description="Whether to enable shaking")
    latchOpen: bool = Field(default=False, description="Whether the latch is open")
    targetSpeed: int = Field(default=1000, description="Target RPM for shaking")
    heaterShakerSetTimer: bool = Field(default=False, description="Whether to set a timer")
    heaterShakerTimerMinutes: int = Field(default=0, description="Timer minutes")
    heaterShakerTimerSeconds: int = Field(default=0, description="Timer seconds") 
    setHeaterShakerTemperature: bool = Field(default=False, description="Whether to set temperature")
    
class ThermocyclerCommand(BaseCommand):
    stepType: str = Field(default="thermocycler", description="The type of the step")
    stepName: str = Field(default="热循环仪", description="Name of the step")
    moduleId: str = Field(..., description="The ID of the thermocycler module")
    lidOpen: bool = Field(default=False, description="Whether the lid is open")
    lidIsActive: bool = Field(default=False, description="Whether the lid temperature control is active")
    blockIsActive: bool = Field(default=False, description="Whether the block temperature control is active")
    profileVolume: str = Field(..., description="Volume for thermocycling")
    lidIsActiveHold: bool = Field(default=False, description="Whether to hold lid temperature")
    profileItemsById: dict = Field(default_factory=dict, description="Dictionary of profile items with either profileStep or profileCycle type")
    blockIsActiveHold: bool = Field(default=False, description="Whether to hold block temperature")
    orderedProfileItems: List[str] = Field(default_factory=list, description="Ordered list of profile item IDs")
    profileTargetLidTemp: str = Field(..., description="Target lid temperature")
    thermocyclerFormType: Literal["thermocyclerProfile", "thermocyclerState"] = Field(default="thermocyclerProfile", description="Type of thermocycler form")
    lidOpenHold: bool = Field(default=False, description="Whether to hold lid open state")
    lidTargetTempHold: str = Field(..., description="Target temperature to hold lid at")
    blockTargetTempHold: str = Field(..., description="Target temperature to hold block at")
    lidTargetTemp: str = Field(..., description="Target lid temperature")
    blockTargetTemp: str = Field(..., description="Target block temperature")

class MagneticModuleCommand(BaseCommand):
    stepType: str = Field(default="magnet", description="The type of the step")
    stepName: str = Field(default="施加磁场", description="Name of the step")
    moduleId: str = Field(..., description="The ID of the magnetic module")
    engageHeight: float = Field(default=0, description="Height to engage")
    magnetAction: Literal["engage", "disengage"] = Field(..., description="Action to take")

class TemperatureCommand(BaseCommand):
    stepType: str = Field(default="temperature", description="The type of the step")
    stepName: str = Field(default="温控", description="Name of the step")
    moduleId: str = Field(..., description="The ID of the temperature module")
    setTemperature: bool = Field(default=False, description="Whether to set temperature")
    targetTemperature: float = Field(..., description="Target temperature in Celsius")

class TransferCommand(BaseCommand):
    stepType: str = Field(default="moveLiquid", description="The type of the step")
    stepName: str = Field(default="移液", description="Name of the step")
    path: str = Field(..., description="Path type for liquid transfer")
    volume: float = Field(..., description="Volume to transfer")
    pipette: str = Field(..., description="ID of pipette to use")
    tipRack: str = Field(..., description="Tip rack to use")
    changeTip: str = Field(..., description="When to change tips")
    preWetTip: bool = Field(default=False, description="Whether to pre-wet tip")
    stepDetails: str = Field(default="", description="Additional step details")
    aspirate_wells: List[str] = Field(..., description="Wells to aspirate from")
    dispense_wells: List[str] = Field(..., description="Wells to dispense to")
    aspirate_labware: str = Field(..., description="Labware to aspirate from")
    dispense_labware: str = Field(..., description="Labware to dispense to")
    dropTip_location: str = Field(..., description="Location to drop tip")
    blowout_checkbox: bool = Field(default=False, description="Whether to perform blowout")
    aspirate_mix_checkbox: bool = Field(default=False, description="Whether to mix while aspirating")
    dispense_mix_checkbox: bool = Field(default=False, description="Whether to mix while dispensing")
    disposalVolume_volume: float = Field(default=30, description="Disposal volume")
    disposalVolume_checkbox: bool = Field(default=False, description="Whether to use disposal volume")
    aspirate_airGap_volume: float = Field(default=30, description="Air gap volume for aspiration")
    aspirate_airGap_checkbox: bool = Field(default=False, description="Whether to use air gap during aspiration")
    dispense_airGap_volume: float = Field(default=30, description="Air gap volume for dispensing")
    dispense_airGap_checkbox: bool = Field(default=False, description="Whether to use air gap during dispensing")
    aspirate_delay_seconds: float = Field(default=1, description="Delay after aspiration")
    aspirate_delay_checkbox: bool = Field(default=False, description="Whether to delay after aspiration")
    dispense_delay_seconds: float = Field(default=1, description="Delay after dispensing")
    dispense_delay_checkbox: bool = Field(default=False, description="Whether to delay after dispensing")
    aspirate_wells_grouped: bool = Field(default=False, description="Whether wells are grouped")
    aspirate_detect_checkbox: bool = Field(default=False, description="Whether to detect liquid level")
    aspirate_wellOrder_first: str = Field(default="t2b", description="First well ordering direction for aspiration")
    aspirate_wellOrder_second: str = Field(default="l2r", description="Second well ordering direction for aspiration")
    dispense_wellOrder_first: str = Field(default="t2b", description="First well ordering direction for dispensing")
    dispense_wellOrder_second: str = Field(default="l2r", description="Second well ordering direction for dispensing")
    aspirate_touchTip_checkbox: bool = Field(default=False, description="Whether to touch tip after aspiration")
    dispense_touchTip_checkbox: bool = Field(default=False, description="Whether to touch tip after dispensing")

class MixCommand(BaseCommand):
    stepType: str = Field(default="mix", description="The type of the step")
    stepName: str = Field(default="混合", description="Name of the step")
    times: int = Field(..., description="Number of mix repetitions")
    wells: List[str] = Field(..., description="Wells to mix in")
    volume: float = Field(..., description="Volume to mix")
    labware: str = Field(..., description="Labware containing wells")
    pipette: str = Field(..., description="ID of pipette to use")
    tipRack: str = Field(..., description="Tip rack to use")
    changeTip: str = Field(..., description="When to change tips")
    stepDetails: str = Field(default="", description="Additional step details")
    blowout_checkbox: bool = Field(default=False, description="Whether to perform blowout")
    dropTip_location: str = Field(..., description="Location to drop tip")
    mix_wellOrder_first: str = Field(default="t2b", description="First well ordering direction")
    mix_wellOrder_second: str = Field(default="l2r", description="Second well ordering direction")
    mix_touchTip_checkbox: bool = Field(default=False, description="Whether to touch tip after mixing")
    aspirate_delay_seconds: float = Field(default=1, description="Delay after aspiration")
    dispense_delay_seconds: float = Field(default=1, description="Delay after dispensing")
    aspirate_delay_checkbox: bool = Field(default=False, description="Whether to delay after aspiration")
    dispense_delay_checkbox: bool = Field(default=False, description="Whether to delay after dispensing")